/*
 * kgraft_patch_bsc973570
 *
 * Fix for bsc#973570
 *
 *  Upstream commit:
 *  a5527dda344fff0514b7989ef7a755729769daa1
 *
 *  SLE12 commit:
 *  9cc7d785fa9868fa71a86cf5a79eb67a28eea039 (3.12.56 -stable)
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <net/sock.h>
#include <net/af_unix.h>

static int (*kgr_unix_mkname)(struct sockaddr_un *sunaddr, int len, unsigned int *hashp);
static int (*kgr_unix_scm_to_skb)(struct scm_cookie *scm, struct sk_buff *skb, bool send_fds);
static struct sock *(*kgr__unix_find_socket_byname)(struct net *net, struct sockaddr_un *sunname, int len, int type, unsigned int hash);
static void (*kgr__unix_insert_socket)(struct hlist_head *list, struct sock *sk);
static struct sock *(*kgr_unix_find_other)(struct net *net, struct sockaddr_un *sunname, int len, int type, unsigned int hash, int *error);
static void (*kgr_unix_dgram_disconnected)(struct sock *sk, struct sock *other);
static void (*kgr_unix_dgram_peer_wake_disconnect)(struct sock *sk, struct sock *other);
static long (*kgr_unix_wait_for_peer)(struct sock *other, long timeo);
static void (*kgr_unix_state_double_lock)(struct sock *sk1, struct sock *sk2);
static int (*kgr_unix_dgram_peer_wake_me)(struct sock *sk, struct sock *other);
static void (*kgr_maybe_add_creds)(struct sk_buff *skb, const struct socket *sock, const struct sock *other);
static void (*kgr_wait_for_unix_gc)(void);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "unix_mkname", (void *)&kgr_unix_mkname },
	{ "unix_scm_to_skb", (void *)&kgr_unix_scm_to_skb },
	{ "__unix_find_socket_byname", (void *)&kgr__unix_find_socket_byname },
	{ "__unix_insert_socket", (void *)&kgr__unix_insert_socket },
	{ "unix_find_other", (void *)&kgr_unix_find_other },
	{ "unix_dgram_disconnected", (void *)&kgr_unix_dgram_disconnected },
	{ "unix_dgram_peer_wake_disconnect", (void *)&kgr_unix_dgram_peer_wake_disconnect },
	{ "unix_wait_for_peer", (void *)&kgr_unix_wait_for_peer },
	{ "unix_state_double_lock", (void *)&kgr_unix_state_double_lock },
	{ "unix_dgram_peer_wake_me", (void *)&kgr_unix_dgram_peer_wake_me },
	{ "maybe_add_creds", (void *)&kgr_maybe_add_creds },
	{ "wait_for_unix_gc", (void *)&kgr_wait_for_unix_gc },
};

#define unix_peer(sk) (unix_sk(sk)->peer)

/* inlined */
static inline unsigned int kgr_unix_hash_fold(__wsum n)
{
	unsigned int hash = (__force unsigned int)csum_fold(n);

	hash ^= hash>>8;
	return hash&(UNIX_HASH_SIZE-1);
}

static void kgr__unix_remove_socket(struct sock *sk)
{
	sk_del_node_init(sk);
}

/* isra optimized */
static int kgr_unix_autobind(struct socket *sock)
{
	struct sock *sk = sock->sk;
	struct net *net = sock_net(sk);
	struct unix_sock *u = unix_sk(sk);
	static u32 ordernum = 1;
	struct unix_address *addr;
	int err;
	unsigned int retries = 0;

	err = mutex_lock_interruptible(&u->readlock);
	if (err)
		return err;

	err = 0;
	if (u->addr)
		goto out;

	err = -ENOMEM;
	addr = kzalloc(sizeof(*addr) + sizeof(short) + 16, GFP_KERNEL);
	if (!addr)
		goto out;

	addr->name->sun_family = AF_UNIX;
	atomic_set(&addr->refcnt, 1);

retry:
	addr->len = sprintf(addr->name->sun_path+1, "%05x", ordernum) + 1 + sizeof(short);
	addr->hash = kgr_unix_hash_fold(csum_partial(addr->name, addr->len, 0));

	spin_lock(&unix_table_lock);
	ordernum = (ordernum+1)&0xFFFFF;

	if (kgr__unix_find_socket_byname(net, addr->name, addr->len, sock->type,
				      addr->hash)) {
		spin_unlock(&unix_table_lock);
		/*
		 * __unix_find_socket_byname() may take long time if many names
		 * are already in use.
		 */
		cond_resched();
		/* Give up if all names seems to be in use. */
		if (retries++ == 0xFFFFF) {
			err = -ENOSPC;
			kfree(addr);
			goto out;
		}
		goto retry;
	}
	addr->hash ^= sk->sk_type;

	kgr__unix_remove_socket(sk);
	u->addr = addr;
	kgr__unix_insert_socket(&unix_socket_table[addr->hash], sk);
	spin_unlock(&unix_table_lock);
	err = 0;

out:	mutex_unlock(&u->readlock);
	return err;
}

/* inlined */
static void kgr_unix_get_secdata(struct scm_cookie *scm, struct sk_buff *skb)
{
	memcpy(UNIXSID(skb), &scm->secid, sizeof(u32));
}

/* inlined */
static inline int kgr_unix_our_peer(struct sock *sk, struct sock *osk)
{
	return unix_peer(osk) == sk;
}

/* inlined */
static inline int kgr_unix_may_send(struct sock *sk, struct sock *osk)
{
	return unix_peer(osk) == NULL || kgr_unix_our_peer(sk, osk);
}

/* inlined */
static void kgr_unix_dgram_peer_wake_disconnect_wakeup(struct sock *sk,
						   struct sock *other)
{
	kgr_unix_dgram_peer_wake_disconnect(sk, other);
	wake_up_interruptible_poll(sk_sleep(sk),
				   POLLOUT |
				   POLLWRNORM |
				   POLLWRBAND);
}

/* inlined */
static inline int kgr_unix_recvq_full(struct sock const *sk)
{
	return skb_queue_len(&sk->sk_receive_queue) > sk->sk_max_ack_backlog;
}

/* patched function */
int kgr_unix_dgram_sendmsg(struct kiocb *kiocb, struct socket *sock,
	struct msghdr *msg, size_t len)
{
	struct sock_iocb *siocb = kiocb_to_siocb(kiocb);
	struct sock *sk = sock->sk;
	struct net *net = sock_net(sk);
	struct unix_sock *u = unix_sk(sk);
	struct sockaddr_un *sunaddr = msg->msg_name;
	struct sock *other = NULL;
	int namelen = 0; /* fake GCC */
	int err;
	unsigned int hash;
	struct sk_buff *skb;
	long timeo;
	struct scm_cookie tmp_scm;
	int max_level;
	int data_len = 0;
	int sk_locked;

	if (NULL == siocb->scm)
		siocb->scm = &tmp_scm;
	kgr_wait_for_unix_gc();
	err = scm_send(sock, msg, siocb->scm, false);
	if (err < 0)
		return err;

	err = -EOPNOTSUPP;
	if (msg->msg_flags&MSG_OOB)
		goto out;

	if (msg->msg_namelen) {
		err = kgr_unix_mkname(sunaddr, msg->msg_namelen, &hash);
		if (err < 0)
			goto out;
		namelen = err;
	} else {
		sunaddr = NULL;
		err = -ENOTCONN;
		other = unix_peer_get(sk);
		if (!other)
			goto out;
	}

	if (test_bit(SOCK_PASSCRED, &sock->flags) && !u->addr
	    && (err = kgr_unix_autobind(sock)) != 0)
		goto out;

	err = -EMSGSIZE;
	if (len > sk->sk_sndbuf - 32)
		goto out;

	if (len > SKB_MAX_ALLOC)
		data_len = min_t(size_t,
				 len - SKB_MAX_ALLOC,
				 MAX_SKB_FRAGS * PAGE_SIZE);

	skb = sock_alloc_send_pskb(sk, len - data_len, data_len,
				   msg->msg_flags & MSG_DONTWAIT, &err,
				   PAGE_ALLOC_COSTLY_ORDER);
	if (skb == NULL)
		goto out;

	err = kgr_unix_scm_to_skb(siocb->scm, skb, true);
	if (err < 0)
		goto out_free;
	max_level = err + 1;
	kgr_unix_get_secdata(siocb->scm, skb);

	skb_put(skb, len - data_len);
	skb->data_len = data_len;
	skb->len = len;
	err = skb_copy_datagram_from_iovec(skb, 0, msg->msg_iov, 0, len);
	if (err)
		goto out_free;

	timeo = sock_sndtimeo(sk, msg->msg_flags & MSG_DONTWAIT);

restart:
	if (!other) {
		err = -ECONNRESET;
		if (sunaddr == NULL)
			goto out_free;

		other = kgr_unix_find_other(net, sunaddr, namelen, sk->sk_type,
					hash, &err);
		if (other == NULL)
			goto out_free;
	}

	if (sk_filter(other, skb) < 0) {
		/* Toss the packet but do not return any error to the sender */
		err = len;
		goto out_free;
	}

	sk_locked = 0;
	unix_state_lock(other);
restart_locked:
	err = -EPERM;
	if (!kgr_unix_may_send(sk, other))
		goto out_unlock;

	if (unlikely(sock_flag(other, SOCK_DEAD))) {
		/*
		 *	Check with 1003.1g - what should
		 *	datagram error
		 */
		unix_state_unlock(other);
		sock_put(other);

		if (!sk_locked)
			unix_state_lock(sk);

		err = 0;
		if (unix_peer(sk) == other) {
			unix_peer(sk) = NULL;
			kgr_unix_dgram_peer_wake_disconnect_wakeup(sk, other);

			unix_state_unlock(sk);

			kgr_unix_dgram_disconnected(sk, other);
			sock_put(other);
			err = -ECONNREFUSED;
		} else {
			unix_state_unlock(sk);
		}

		other = NULL;
		if (err)
			goto out_free;
		goto restart;
	}

	err = -EPIPE;
	if (other->sk_shutdown & RCV_SHUTDOWN)
		goto out_unlock;

	if (sk->sk_type != SOCK_SEQPACKET) {
		err = security_unix_may_send(sk->sk_socket, other->sk_socket);
		if (err)
			goto out_unlock;
	}

	/*
	 * Fix bsc#973570
	 *  -if (unlikely(unix_peer(other) != sk && kgr_unix_recvq_full(other))) {
	 *  +next 6 rows
	 */
	/* other == sk && unix_peer(other) != sk if
	 * - unix_peer(sk) == NULL, destination address bound to sk
	 * - unix_peer(sk) == sk by time of get but disconnected before lock
	 */
	if (other != sk &&
	    unlikely(unix_peer(other) != sk && kgr_unix_recvq_full(other))) {
		if (timeo) {
			timeo = kgr_unix_wait_for_peer(other, timeo);

			err = sock_intr_errno(timeo);
			if (signal_pending(current))
				goto out_free;

			goto restart;
		}

		if (!sk_locked) {
			unix_state_unlock(other);
			kgr_unix_state_double_lock(sk, other);
		}

		if (unix_peer(sk) != other ||
		    kgr_unix_dgram_peer_wake_me(sk, other)) {
			err = -EAGAIN;
			sk_locked = 1;
			goto out_unlock;
		}

		if (!sk_locked) {
			sk_locked = 1;
			goto restart_locked;
		}
	}

	if (unlikely(sk_locked))
		unix_state_unlock(sk);

	if (sock_flag(other, SOCK_RCVTSTAMP))
		__net_timestamp(skb);
	kgr_maybe_add_creds(skb, sock, other);
	skb_queue_tail(&other->sk_receive_queue, skb);
	if (max_level > unix_sk(other)->recursion_level)
		unix_sk(other)->recursion_level = max_level;
	unix_state_unlock(other);
	other->sk_data_ready(other, len);
	sock_put(other);
	scm_destroy(siocb->scm);
	return len;

out_unlock:
	if (sk_locked)
		unix_state_unlock(sk);
	unix_state_unlock(other);
out_free:
	kfree_skb(skb);
out:
	if (other)
		sock_put(other);
	scm_destroy(siocb->scm);
	return err;
}

static int kgr_patch_bsc973570_kallsyms(void)
{
        unsigned long addr;
        int i;

        for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
                addr = kallsyms_lookup_name(kgr_funcs[i].name);
                if (!addr) {
                        pr_err("kgraft-patch: symbol %s not resolved\n",
                                kgr_funcs[i].name);
                        return -ENOENT;
                }

                *(kgr_funcs[i].addr) = (void *)addr;
        }

        return 0;
}

int kgr_patch_bsc973570_init(void)
{
        return kgr_patch_bsc973570_kallsyms();
}
